/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Run
Version:   1.0
ID:        ID_RUN
Category:  Command
Date:      September 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Short:     The execute class runs native binary programs.
Keywords:  execute, executable, run, binary, program, task

###DESCRIPTION###
<p>The Run class allows you to launch binary files that are stored in the
file system.  Executable files must be native to the system that Athene is
running on.  For example, if you are running Athene on a Linux system, the
Run class will allow you to run X11 programs and popular applications such
as Netscape.</p>

<p>The Run class is not restricted to launching executable files.  For
instance, you can launch picture, text and custom file types.  The exact
behaviour that occurs when launching a data file is dependent on the
settings found in the user's &quot;user:config/associations.cfg&quot;
file, which holds the launch settings for each registered class and file
type.  In most cases a script will be run which can handle the file
data.</p>

<p>Typical usage of the Run class in a DML script is demonstrated in the
following examples:</p>

<pre>
   &lt;run src="netscape"/&gt;
   &lt;run mode="view" src="pictures:screenshot.png"/&gt;
</pre>

<p>If you need to send arguments to an executable file, use the
<a href="#Args">Args</a> field as in this example:</p>

<pre>
   &lt;run src="myprogram" args="arg1 arg2 arg3"/&gt;
</pre>

<p>Like all commands, you may make a run object static if you want it
to be activated in response to system events.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/data/config.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_RUN 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_RUN, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "September 2003",
   "Run"
};

static OBJECTPTR RunClass = 0;
static OBJECTPTR modFileSystem = 0;
static OBJECTPTR modStrings = 0;
static struct KernelBase *KernelBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static struct StringsBase *StringsBase = 0;

/*****************************************************************************
** Run class definition.
*/

struct Run {
   OBJECT_HEADER        /* Class header */
   LONG     Static;     /* Set if the Run object should reside in memory */
   OBJECTID OutputID;   /* Reference to an output object */
   LONG     Mode;       /* The type of action to perform in launching the file */
   LONG     Flags;
   LONG     ReturnCode;

   PRIVATE_FIELDS
   UBYTE  Buffer[1024];   /* The buffer is required in the structure to prevent forking problems */
   STRING Location;       /* The location of the program */
   STRING Args;           /* The arguments to pass to the program */
   MEMORYID LocationMID;
   MEMORYID ArgsMID;
};

#define RNL_OPEN  1
#define RNL_EDIT  2
#define RNL_PRINT 3
#define RNL_VIEW  4

#define RNF_FOREIGN  0x00000001
#define RNF_WAIT     0x00000002
#define RNF_RESETDIR 0x00000004

static ERROR GET_Args(struct Run *, STRING *);
static ERROR GET_Location(struct Run *, STRING *);

static ERROR SET_Args(struct Run *, STRING);
static ERROR SET_Location(struct Run *, STRING);

static struct FieldDef ModeLookup[] = {
   { "OPEN",  RNL_OPEN },
   { "EDIT",  RNL_EDIT },
   { "PRINT", RNL_PRINT },
   { "VIEW",  RNL_VIEW },
   { NULL, NULL }
};

static struct FieldDef RunFlags[] = {
   { "FOREIGN",  RNF_FOREIGN },
   { "WAIT",     RNF_WAIT },
   { "RESETDIR", RNF_RESETDIR },
   { NULL, NULL }
};

static struct FieldArray RunFields[] = {
  { "Static",     0, FDF_LONG|FDF_RI,            0, NULL, NULL },
  { "Output",     0, FDF_OBJECTID|FDF_RI,        0, NULL, NULL },
  { "Mode",       0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&ModeLookup, NULL, NULL },
  { "Flags",      0, FDF_LONGFLAGS|FDF_RI,       (LONG)&RunFlags, NULL, NULL },
  { "ReturnCode", 0, FDF_LONG|FDF_R,             0, NULL, NULL },
  /*** Virtual fields ***/
  { "Args",     0, FDF_STRING|FDF_RW,   0, GET_Args,     SET_Args },
  { "Location", 0, FDF_STRING|FDF_RW,   0, GET_Location, SET_Location },
  { "Src",      0, FDF_STRING|FDF_RW,   0, GET_Location, SET_Location },
  END_FIELD
};

static ERROR RUN_Activate(struct Run *, APTR);
static ERROR RUN_ClosingTag(struct Run *, APTR);
static ERROR RUN_Free(struct Run *, APTR);
static ERROR RUN_ReleaseObject(struct Run *, APTR);
static ERROR RUN_Init(struct Run *, APTR Void);

static struct ActionArray RunActions[] = {
   { AC_Activate,      RUN_Activate },
   { AC_ClosingTag,    RUN_ClosingTag },
   { AC_Free,          RUN_Free },
   { AC_ReleaseObject, RUN_ReleaseObject },
   { AC_Init,          RUN_Init },
   { NULL, NULL }
};

static void Print(struct Run *, STRING);
static ERROR TestClass(struct Run *, OBJECTPTR, OBJECTPTR, CLASSID, STRING *);

/*****************************************************************************
** User Instructions.
*/

static BYTE RunUsage[] = {
"USAGE: RUN SRC=filename ARGS=\"arguments...\"\n\
\n\
Executable and data files can be run with this command.  For executable\n\
files, the file must be in a format that is native to the system that\n\
Athene is running on.  Arguments can be passed to executable programs\n\
via the Args string.  Data files can be in any format that is recognised\n\
by your \"user:config/associations.cfg\" file.\n\
\n\
Here are some examples:\n\
\n\
  run src=\"xterm\"\n\
  run src=\"demos:boing.dml\"\n\
  run src=\"pictures:screenshot.png\"\n\
  run src=\"bin:runscript\" args=\"-script templates:window.dml -wait\"\n"
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** File System module ***/

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Strings module ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&RunClass, NULL,
      FID_BaseClassID|TLONG, ID_RUN,
      FID_Version|TFLOAT,    VER_RUN,
      FID_Name|TSTRING,      "Run",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        RunUsage,
      FID_Actions|TPTR,      RunActions,
      FID_Fields|TPTR,       RunFields,
      FID_Size|TLONG,        sizeof(struct Run),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modStrings)    { acFree(modStrings);    modStrings = 0; }
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = 0; }
   if (RunClass)      { acFree(RunClass);      RunClass = 0; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates a Run object.
###END###

*****************************************************************************/

static ERROR RUN_Activate(struct Run *Self, APTR Void)
{
   UBYTE buffer[800], commandfile[200];
   struct mtReadConfig read;
   struct ConfigEntry *entries;
   STRING location, args, command, str, mode;
   OBJECTPTR task, classes, config;
   LONG ClassID, SubClassID, i, j, amtentries, ext, flags, type;
   OBJECTID SystemClassesID;

   /*** Get the location and argument strings ***/

   if (!Self->LocationMID) {
      Print(Self, "The location of the executable was not specified.");
      return(ERR_FieldNotSet);
   }

   if (GET_Location(Self, &location) != ERR_Okay) return(ObjectError(Self, ERH_Activate, ERR_GetField));

   /*** Check that the location refers to a valid file ***/

   if ((AnalyseLocation(location, &type) != ERR_Okay) OR (type != LOC_FILE)) {
      return(ERR_FileNotFound);
   }

   command = NULL;
   str = NULL;

   /*** Determine the start of the file extension, if there is one ***/

   for (ext=0; location[ext]; ext++);
   while ((ext > 0) AND (location[ext] != '.') AND (location[ext] != ':') AND (location[ext] != '/') AND (location[ext] != '\\')) ext--;
   if (location[ext] IS '.') ext++;
   else ext = 0;

   if (Self->Mode IS RNL_EDIT)       mode = "Edit";
   else if (Self->Mode IS RNL_OPEN)  mode = "Open";
   else if (Self->Mode IS RNL_PRINT) mode = "Print";
   else if (Self->Mode IS RNL_VIEW)  mode = "View";
   else mode = "Open";

   /* Determine the class that the file belongs to, or the assocation
   ** that should be used for executing the file.
   */

   for (i=0; location[i]; i++) if ((location[i] IS ':') OR (location[i] IS '/') OR (location[i] IS '\\')) break;
   if (!location[i]) ClassID = ID_TASK;
   else {
      /* Note that if the IdentifyFile() function fails it will be
      ** because the data does not belong to a specific class.
      ** The associations.cfg file will help us load files that
      ** do not have class associations later in this subroutine.
      */
      IdentifyFile(location, &ClassID, &SubClassID);
   }

   /*** If the file is an executable, run it immediately and return ***/

   if ((ClassID IS ID_TASK) OR (Self->Flags & RNF_FOREIGN)) {

      Self->ReturnCode = -1;

      GET_Args(Self, &args);

      flags = NULL;

      if (Self->Flags & RNF_FOREIGN) flags |= TSF_FOREIGN;
      if (Self->Flags & RNF_WAIT) flags |= TSF_WAIT;
      if (Self->Flags & RNF_RESETDIR) flags |= TSF_RESETDIR;

      if (CreateObject(ID_TASK, NF_CHILD, &task, NULL,
         FID_Location|TSTRING, location,
         FID_Output|TLONG,     Self->OutputID,
         FID_Args|TSTRING,     args,
         FID_Flags|TLONG,      flags,
         TAGEND) IS ERR_Okay) {

         if (Action(AC_Activate, task, NULL) IS ERR_Okay) {

            if (Self->Flags & RNF_WAIT) {
               GetField(task, FID_ReturnCode, FT_LONG, &Self->ReturnCode);
               DPrintF("Run:","Return Code: %d", Self->ReturnCode);
            }
            else Self->ReturnCode = 0;

            Action(AC_Free, task, NULL);
            return(ERR_Okay);
         }
         else {
            Action(AC_Free, task, NULL);
            Print(Self, "The executable program could not be loaded.");
            return(ERR_Activate);
         }
      }
      else {
         Print(Self, "Internal failure (A Task object could not be created).");
         return(ERR_CreateObject);
      }
   }

   /*** Load the associations configuration file ***/

   if (CreateObject(ID_CONFIG, NULL, &config, NULL,
      FID_Location|TSTRING, "system:config/software/associations.cfg;user:config/associations.cfg",
      TAGEND) != ERR_Okay) {
      Print(Self, "Failed to load the \"user:config/associations.cfg\" file.");
      return(ERR_CreateObject);
   }

   if (ClassID) {
      /* Get the name of the class and sub-class so that we can search
      ** the SystemAssociations object.
      */

      if (FastFindObject("SystemClasses", NULL, &SystemClassesID, 1, NULL) IS ERR_Okay) {
         if (AccessObject(SystemClassesID, 3000, &classes) IS ERR_Okay) {
            if ((!SubClassID) OR (TestClass(Self, classes, config, SubClassID, &command) != ERR_Okay)) {
               TestClass(Self, classes, config, ClassID, &command);
            }
            ReleaseObject(classes);
         }
      }
   }

   /* If no class association exists for this file or if a command
   ** reference could not be yielded, search the associations
   ** file to see if the file extension (e.g. *.txt) has a command
   ** assocation.
   */

   if ((!command) AND (location[ext])) {
      if (GetFields(config, FID_AmtEntries|TLONG, &amtentries,
                            FID_Entries|TPTR,     &entries,
                            TAGEND) IS ERR_Okay) {
         for (i=0; i < amtentries; i++) {
            if (StrCompare(entries[i].Item, "extension", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               if (StrCompare(entries[i].Data, location+ext, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  read.Section = entries[i].Section;
                  read.Item    = mode;
                  if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) command = read.Data;
               }
            }
         }
      }
   }

   /* If no association exists for the file then use the default
   ** options if they are available.
   */

   if (!command) {
      read.Section = "default";
      read.Item = mode;
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) command = read.Data;
   }

   /*** Process the command string (must either be a binary or a script) ***/

   if (!command) {
      Print(Self, "Failed to identify the file's type.");
      Action(AC_Free, config, NULL);
      return(ERR_Failed);
   }

   /*** Extract the script or binary that we are asked to run ***/

   j = 0;
   commandfile[0] = 0;
   if (*command IS '"') {
      command++;
      while ((*command) AND (*command != '"') AND (i < sizeof(commandfile)-1)) commandfile[j++] = *command++;
      if (*command IS '"') command++;
   }
   else while ((*command) AND (*command != ' ') AND (i < sizeof(commandfile)-1)) commandfile[j++] = *command++;
   commandfile[j] = 0;

   while (*command IS ' ') command++;

   /*** Test the command reference to see if it is a script or a program and respond accordingly ***/

   ClassID = NULL;
   IdentifyFile(commandfile, &ClassID, NULL);

   if (ClassID IS ID_TASK) StrCopy(command, buffer, sizeof(buffer));
   else StrFormat(buffer, sizeof(buffer), "-target desktop -script \"%s\" -args %s", commandfile, command);

   /*** Free the config object as we no longer need it ***/

   Action(AC_Free, config, NULL);

   /* Convert any references to {location} to the location of
   ** the file that we are attempting to run.
   */

   if (StrReplace(buffer, "{location}", location, &str, FALSE) IS ERR_Okay) {
      StrCopy(str, buffer, sizeof(buffer));
      FreeMemory(str);
   }

   /*** Launch the command ***/

   if (ClassID IS ID_TASK) {
      CreateObject(ID_TASK, NF_CHILD, &task, NULL,
         FID_Output|TLONG,     Self->OutputID,
         FID_Location|TSTRING, commandfile,
         FID_Args|TSTRING,     buffer,
         TAGEND);
   }
   else {
      CreateObject(ID_TASK, NF_CHILD, &task, NULL,
         FID_Location|TSTRING, "bin:runscript",
         FID_Output|TLONG,     Self->OutputID,
         FID_Args|TSTRING,     buffer,
         TAGEND);
   }

   if (!task) {
      Print(Self, "Internal failure (A Task object could not be created).");
      return(ERR_CreateObject);
   }

   if (Action(AC_Activate, task, NULL) IS ERR_Okay) {
      Action(AC_Free, task, NULL);
      return(ERR_Okay);
   }
   else {
      Action(AC_Free, task, NULL);
      Print(Self, "The executable program could not be loaded.");
      return(ERR_Activate);
   }
}

/*****************************************************************************

###ACTION###
Name:  ClosingTag
Short: If the object is non-static, this action will start the execution process and then the object will self-destruct.
###END###

*****************************************************************************/

static ERROR RUN_ClosingTag(struct Run *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
      Action(AC_Free, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Run: Free
*/

static ERROR RUN_Free(struct Run *Self, APTR Void)
{
   if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location = NULL; }
   if (Self->Args)        { ReleaseMemory(Self->Args);       Self->Args = NULL; }
   if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }
   if (Self->ArgsMID)     { FreeMemoryID(Self->ArgsMID);     Self->ArgsMID = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Run: Init
*/

static ERROR RUN_Init(struct Run *Self, APTR Void)
{
   if (!Self->LocationMID) return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));

   if (!Self->Mode) Self->Mode = RNL_OPEN;

   return(ERR_Okay);
}

/*****************************************************************************
** Run: ReleaseObject
*/

static ERROR RUN_ReleaseObject(struct Run *Self, APTR Void)
{
   if (Self->Location) { ReleaseMemory(Self->Location); Self->Location = NULL; }
   if (Self->Args)     { ReleaseMemory(Self->Args);     Self->Args = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Args
Short:  Set this field to send the executable program a range of arguments.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>If the executable program accepts arguments, then you may send them to
the program by setting this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Args(struct Run *Self, STRING *Value)
{
   if (Self->Args) {
      *Value = Self->Args;
      return(ERR_Okay);
   }
   else if (!Self->ArgsMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ArgsMID, MEM_READ, (void **)&Self->Args) IS ERR_Okay) {
      *Value = Self->Args;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Args(struct Run *Self, STRING Value)
{
   LONG i;

   if (Self->Args)    { ReleaseMemory(Self->Args);   Self->Args = NULL; }
   if (Self->ArgsMID) { FreeMemoryID(Self->ArgsMID); Self->ArgsMID = NULL; }

   if ((Value) AND (*Value)) {
      if ((Value[0] IS '1') AND (!Value[1])) return(ERR_Okay); /* Ignore arguments of "1" */
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Args, &Self->ArgsMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Args[i] = Value[i];
         Self->Args[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Defines special options to use when running programs.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that affect the behaviour of the Run class may be specified in this field.</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">FOREIGN</td><td>If the program that you are running is not native to Athene or Pandora then you must set the FOREIGN flag so that the system does not attempt to treat it as a native program.</td></tr>
<tr><td width="10%">RESETDIR&nbsp;&nbsp;</td><td>Normally when a program is launched, it will start in the same directory as your own process.  If you would rather that the program starts in the same directory as its executable, set the RESETDIR flag.</td></tr>
<tr><td width="10%">WAIT</td><td>If this option is set then your task will be stop and wait for the other program to finish when launched.  This can be very useful when running batch scripts.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Defines the location/name of the executable program.
Type:     STRING
Status:   Read/Write

###DESCRIPTION###
<p>It is compulsory that this field is set to the location of the
executable program that you wish to run.  Directory assigns are fully
supported, so you may use this file-system feature in specifying the
location.</p>

<p>If you are running Athene on a Unix system then you should know that the
PATH variable is supported, so you do not need to fully-qualify the location
if it is in the path.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct Run *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else if (!Self->LocationMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->LocationMID, MEM_READ, (void **)&Self->Location) IS ERR_Okay) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Location(struct Run *Self, STRING Value)
{
   STRING path;
   LONG i;

   if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location = NULL; }
   if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }

   if ((Value) AND (*Value)) {
      if (ResolveAssign(Value, &path, RSF_APPROXIMATE) IS ERR_Okay) {
         for (i=0; path[i]; i++);
         if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Location, &Self->LocationMID) IS ERR_Okay) {
            for (i=0; path[i]; i++) Self->Location[i] = path[i];
            Self->Location[i] = 0;
            FreeMemory(path);
            return(ERR_Okay);
         }
         else {
            FreeMemory(path);
            return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
         }
      }
      else {
         DPrintF("@Set.Location:","[Run] Invalid location \"%s\"", Value);
         return(ERR_ResolveAssign);
      }
   }
   else return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Mode
Short:  The type of launch mode to use when launching data files.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The Mode field determines the type of action that will be performed when
running a data file.  The default mode is 'Open', which will open the
data file in the preferred user mode (in most cases a viewing application
will be loaded for this purpose).  Other mode types that are available to
you are listed in the following table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Value</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">EDIT</td><td>Edit mode is used for running an application that can edit the data file.</td></tr>
<tr><td width="10%">OPEN</td><td>Open mode will open the file in a way that is appropriate for the data in question (for instance, opening an audio file will play the audio data).</td></tr>
<tr><td width="10%">PRINT&nbsp;&nbsp;</td><td>Print mode is used to instantly print a data file to the default printer.  Printing can only be used on data files that have some form of visual representation.</td></tr>
<tr><td width="10%">VIEW</td><td>View mode is used to view the file data in the user display.  This mode can only be used on data files that have some form of visual representation.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make the object static.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>By default, a run object will activate itself and then self-destruct
when a closing tag is received.  If you would rather that the object stays in
the system, set this field to TRUE.  If you do this, the only way to get the
run object to perform is to call the Activate() action.</p>

###END###

*****************************************************************************/

/*****************************************************************************
** Internal: Print()
*/

static void Print(struct Run *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      channel.ObjectID = Self->Head.UniqueID;
      channel.DataType = DATA_TEXT;
      channel.Version  = 1;
      channel.Buffer   = Buffer;
      for (channel.Size=0; Buffer[channel.Size]; channel.Size++);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
   else DPrintF("Run:", Buffer);
}

/*****************************************************************************
** Internal: TestClass()
*/

static ERROR TestClass(struct Run *Self, OBJECTPTR Config, OBJECTPTR Associations, CLASSID ClassID, STRING *Script)
{
   #undef LEN_CLASSNAME
   #define LEN_CLASSNAME 30
   BYTE ClassName[LEN_CLASSNAME];
   struct mtReadConfig read;
   struct ConfigEntry *entries;
   LONG amtentries, i, j;

   if ((!ClassID) OR (!Config) OR (!Script)) return(SendError(ERH_Function, ERR_Args));

   if (GetFields(Config, FID_AmtEntries|TLONG, &amtentries,
                         FID_Entries|TPTR,     &entries,
                         TAGEND) IS ERR_Okay) {

      for (i=0; i < amtentries; i++) {
         if (StrCompare(entries[i].Item, "ClassID", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            if (ClassID IS StrToInt(entries[i].Data)) {
               for (j=0; (entries[i].Section[j]) AND (j < LEN_CLASSNAME-1); j++) ClassName[j] = entries[i].Section[j];
               ClassName[j] = 0;

               if (GetFields(Associations, FID_AmtEntries|TLONG, &amtentries,
                                           FID_Entries|TPTR,     &entries,
                                           TAGEND) IS ERR_Okay) {

                  for (i=0; i < amtentries; i++) {
                     if (StrCompare(entries[i].Item, "Class", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (StrCompare(entries[i].Data, ClassName, NULL, STR_MATCHLENGTH) IS ERR_Okay) {

                           read.Section = entries[i].Section;
                           if (Self->Mode IS RNL_EDIT)       read.Item = "Edit";
                           else if (Self->Mode IS RNL_OPEN)  read.Item = "Open";
                           else if (Self->Mode IS RNL_PRINT) read.Item = "Print";
                           else if (Self->Mode IS RNL_VIEW)  read.Item = "View";
                           else read.Item = "Open";

                           if (Action(MT_ReadConfig, Associations, &read) IS ERR_Okay) *Script = read.Data;
                           return(ERR_Okay);
                        }
                     }
                  }
               }
               else return(SendError(ERH_Function, ERR_GetField));

               return(ERR_Search);
            }
         }
      }
      return(ERR_Search);
   }
   else return(SendError(ERH_Function, ERR_GetField));
}
