/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Input
Version:   1.0
ID:        ID_INPUT
Status:    Beta
Category:  GUI
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003-2004.  All rights reserved.
Keywords:  input, text, entry, box
Short:     The Input class manages the display and interactivity of user input boxes.

###DESCRIPTION###
<p>The Input class simplifies the creation and management of input boxes as
part of the user interface.  New input areas can be created by specifying as
little as the graphical dimensions for the box area.  The Input class allows
for the specifics of the graphics to be altered, such as the colours and the
font used.</p>

<p>The definitions for new input boxes are loaded by default from the
environment file "templates:inputdef.xml".  You can change the template file
prior to initialisation by setting the Template field.  Note that any values
set in the template will override your original field settings for the
input object.</p>

<p>You may need to set the object up so that when the user clicks or tabs
away from the input box, it performs an action.  There are three accepted
methods that you can use to achieve this.  The first is to set the
ActionScript field so that it refers to a script that executes when the
input box is activated.  The second is to initialise child objects to the
input object and they will be executed on activation. The third option is to
listen to the Activate action by calling the SubscribeAction() function on
the input.</p>

<p>The following example illustrates how you might create an input box in
DML.  Notice that the Set object is initialised as static and will be
executed when the input object is activated:</p>

<pre>
  &lt;input xoffset="[window.leftmargin]" yoffset="[window.bottommargin]"
    text="Hello World" width="70"&gt;
    &lt;set static object="[text]" &string="[owner.text]"/&gt;
  &lt;/input&gt;
</pre>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_INPUT  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_INPUT, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003-2004. All rights reserved.",
   "January 2004",
   "Input"
};

static struct KernelBase  *KernelBase;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR InputClass = NULL;
static OBJECTPTR StringsMod = NULL;
static OBJECTPTR FileSystemMod = NULL;

enum {
   STATE_ENTERED=1,
   STATE_EXITED,
   STATE_INSIDE
};

/*****************************************************************************
** Object structure.
*/

struct Input {
   OBJECT_HEADER              /* Class header */
   OBJECTID RegionID;         /* Drawable region created by the input object */
   OBJECTID DrawableID;       /* The drawable target for the input graphic */
   LONG     Flags;            /* Special options */
   struct   Font *Font;       /* Font control object - for the label graphic */
   OBJECTPTR Text;            /* Text control object - for the input area */
   LONG     EnterFrame;       /* Frame to display when the mouse enters the input area */
   LONG     ExitFrame;        /* Frame to revert to when the mouse leaves the input area */
   LONG     FocusFrame;       /* Frame to display when the input box has the focus */
   LONG     ReleaseFrame;     /* Frame to display when the input box loses the focus */
   LONG     Thickness;        /* Border thickness */
   LONG     LabelWidth;       /* Width of the input box label area */

   PRIVATE_FIELDS
   struct RGB ColourRGB;      /* Colour to use as the background */
   struct RGB HighlightRGB;   /* Colour to use for border highlighting */
   struct RGB ShadowRGB;      /* Colour to use for border shadowing */
   UBYTE     Label[50];       /* Label */
   OBJECTPTR TemplateXML;
   OBJECTID  TextID;
   OBJECTID  FontID;
   MEMORYID  ActionScriptMID;
   STRING    ActionScript;     /* Simple DML statement indicating what to do on input-press, e.g. "<set object="[desktop]" field="">"... */
   BYTE      State;
   BYTE      Active;
};

/*****************************************************************************
** Class definition.
*/

static struct FieldDef InputFlags[] = {
   { "HIDE",          INF_HIDE },
   { "DISABLED",      INF_DISABLED },
   { "NOBKGD",        INF_NOBKGD },
   { "LOCAL",         INF_LOCAL },
   { "SUNKEN",        INF_SUNKEN },
   { "NOFOCUS",       INF_NOFOCUS },
   { "RAISED",        INF_RAISED },
   { "COMMANDLINE",   INF_COMMANDLINE },
   { "FOCUSACTIVATE", INF_FOCUSACTIVATE },
   { "SELECTTEXT",    INF_SELECTTEXT },
   { "SECRET",        INF_SECRET },
   { NULL, NULL }
};

static ERROR GET_ActionScript(struct Input *, STRING *);
static ERROR GET_Bottom(struct Input *, LONG *);
static ERROR GET_ColourRGB(struct Input *, struct RGB *);
static ERROR GET_Disable(struct Input *, LONG *);
static ERROR GET_FocusActivate(struct Input *, LONG *);
static ERROR GET_Height(struct Input *, struct Variable *);
static ERROR GET_HighlightRGB(struct Input *, struct RGB *);
static ERROR GET_Label(struct Input *, STRING *);
static ERROR GET_Raised(struct Input *, LONG *);
static ERROR GET_Right(struct Input *, LONG *);
static ERROR GET_ShadowRGB(struct Input *, struct RGB *);
static ERROR GET_String(struct Input *, STRING *);
static ERROR GET_Sunken(struct Input *, LONG *);
static ERROR GET_Width(struct Input *, struct Variable *);
static ERROR GET_X(struct Input *, struct Variable *);
static ERROR GET_XOffset(struct Input *, struct Variable *);
static ERROR GET_Y(struct Input *, struct Variable *);
static ERROR GET_YOffset(struct Input *, struct Variable *);

static ERROR SET_ActionScript(struct Input *, STRING);
static ERROR SET_Colour(struct Input *, STRING);
static ERROR SET_Disable(struct Input *, LONG *);
static ERROR SET_FocusActivate(struct Input *, LONG *);
static ERROR SET_Height(struct Input *, struct Variable *);
static ERROR SET_Highlight(struct Input *, STRING);
static ERROR SET_Label(struct Input *, STRING);
static ERROR SET_LabelWidth(struct Input *, LONG *);
static ERROR SET_Raised(struct Input *, LONG *);
static ERROR SET_Region(struct Input *, LONG *);
static ERROR SET_Shadow(struct Input *, STRING);
static ERROR SET_String(struct Input *, STRING);
static ERROR SET_Sunken(struct Input *, LONG *);
static ERROR SET_TabFocus(struct Input *, OBJECTID *);
static ERROR SET_Template(struct Input *, STRING);
static ERROR SET_Width(struct Input *, struct Variable *);
static ERROR SET_X(struct Input *, struct Variable *);
static ERROR SET_XOffset(struct Input *, struct Variable *);
static ERROR SET_Y(struct Input *, struct Variable *);
static ERROR SET_YOffset(struct Input *, struct Variable *);

static struct FieldArray InputFields[] = {
   { "Region",         0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, SET_Region },
   { "Drawable",       0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RW, (LONG)&InputFlags, NULL, NULL },
   { "Font",           0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "InputText",      0, FDF_CHILD|FDF_R,     0, NULL, NULL },
   { "EnterFrame",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ExitFrame",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "FocusFrame",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ReleaseFrame",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Thickness",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "LabelWidth",     0, FDF_LONG|FDF_RW,      0, NULL, SET_LabelWidth },
   /** Virtual fields ***/
   { "ActionScript",   0, FDF_STRING|FDF_RW,   0, GET_ActionScript,  SET_ActionScript },
   { "Bottom",         0, FDF_LONG|FDF_R,      0, GET_Bottom,        NULL },
   { "Disable",        0, FDF_LONG|FDF_RW,     0, GET_Disable,       SET_Disable },
   { "Colour",         0, FDF_STRING|FDF_W,    0, NULL,              SET_Colour },
   { "ColourRGB",      0, FDF_POINTER|FDF_R,   0, GET_ColourRGB,     NULL },
   { "FocusActivate",  0, FDF_LONG|FDF_RW,     0, GET_FocusActivate, SET_FocusActivate },
   { "Highlight",      0, FDF_STRING|FDF_W,    0, NULL,              SET_Highlight },
   { "HighlightRGB",   0, FDF_POINTER|FDF_R,   0, GET_HighlightRGB,  NULL },
   { "Label",          0, FDF_STRING|FDF_RW,   0, GET_Label,         SET_Label },
   { "Raised",         0, FDF_LONG|FDF_RW,     0, GET_Raised,        SET_Raised },
   { "Right",          0, FDF_LONG|FDF_R,      0, GET_Right,         NULL },
   { "Shadow",         0, FDF_STRING|FDF_W,    0, NULL,              SET_Shadow },
   { "ShadowRGB",      0, FDF_POINTER|FDF_R,   0, GET_ShadowRGB,     NULL },
   { "Sunken",         0, FDF_LONG|FDF_RW,     0, GET_Sunken,        SET_Sunken },
   { "String",         0, FDF_STRING|FDF_RW,   0, GET_String,        SET_String },
   { "TabFocus",       0, FDF_OBJECTID|FDF_W,  ID_TABFOCUS, NULL,    SET_TabFocus },
   { "Template",       0, FDF_STRING|FDF_I,    0, NULL,              SET_Template },
   { "Text",           0, FDF_STRING|FDF_RW,   0, GET_String,        SET_String },
   /*** Variable Fields ***/
   { "Height",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR INPUT_AccessObject(struct Input *, APTR);
static ERROR INPUT_ActionNotify(struct Input *, struct acActionNotify *);
static ERROR INPUT_Activate(struct Input *, APTR);
static ERROR INPUT_ClosingTag(struct Input *, APTR);
static ERROR INPUT_DataChannel(struct Input *, struct acDataChannel *);
static ERROR INPUT_Disable(struct Input *, APTR);
static ERROR INPUT_Enable(struct Input *, APTR);
static ERROR INPUT_Focus(struct Input *, APTR);
static ERROR INPUT_Free(struct Input *, APTR);
static ERROR INPUT_Hide(struct Input *, APTR);
static ERROR INPUT_Init(struct Input *, APTR);
static ERROR INPUT_MoveToBack(struct Input *, APTR);
static ERROR INPUT_MoveToFront(struct Input *, APTR);
static ERROR INPUT_NewObject(struct Input *, APTR);
static ERROR INPUT_Redimension(struct Input *, struct acRedimension *);
static ERROR INPUT_ReleaseObject(struct Input *, APTR);
static ERROR INPUT_Resize(struct Input *, struct acResize *);
static ERROR INPUT_Show(struct Input *, APTR);

static struct ActionArray InputActions[] = {
   { AC_AccessObject,  INPUT_AccessObject },
   { AC_ActionNotify,  INPUT_ActionNotify },
   { AC_Activate,      INPUT_Activate },
   { AC_ClosingTag,    INPUT_ClosingTag },
   { AC_DataChannel,   INPUT_DataChannel },
   { AC_Disable,       INPUT_Disable },
   { AC_Enable,        INPUT_Enable },
   { AC_Focus,         INPUT_Focus },
   { AC_Free,          INPUT_Free },
   { AC_Hide,          INPUT_Hide },
   { AC_Init,          INPUT_Init },
   { AC_MoveToBack,    INPUT_MoveToBack },
   { AC_MoveToFront,   INPUT_MoveToFront },
   { AC_NewObject,     INPUT_NewObject },
   { AC_Redimension,   INPUT_Redimension },
   { AC_ReleaseObject, INPUT_ReleaseObject },
   { AC_Resize,        INPUT_Resize },
   { AC_Show,          INPUT_Show },
   { NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

static void DrawInput(struct Input *, struct Render *, struct Bitmap *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&InputClass, NULL,
      FID_BaseClassID|TLONG, ID_INPUT,
      FID_Version|TFLOAT,    VER_INPUT,
      FID_Name|TSTRING,      "Input",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN,
      FID_Actions|TPTR,      InputActions,
      FID_Fields|TPTR,       InputFields,
      FID_Size|TLONG,        sizeof(struct Input),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (InputClass)   { Action(AC_Free, InputClass, NULL);   InputClass = NULL; }
   if (StringsMod)    { Action(AC_Free, StringsMod, NULL);    StringsMod = NULL; }
   if (FileSystemMod) { Action(AC_Free, FileSystemMod, NULL); FileSystemMod = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Input: AccessObject
*/

static ERROR INPUT_AccessObject(struct Input *Self, APTR Void)
{
   if (Self->FontID) {
      if (AccessObject(Self->FontID, 0, (OBJECTPTR *)&Self->Font) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }

   if (Self->TextID) {
      if (AccessObject(Self->TextID, 0, (OBJECTPTR *)&Self->Text) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Input: ActionNotify
*/

static ERROR INPUT_ActionNotify(struct Input *Self, struct acActionNotify *Args)
{
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS Self->TextID)) {
      Action(AC_Activate, Self, NULL);
   }
   else if (Args->ActionID IS AC_UserMovement) {
      struct acUserMovement *movement;
      OBJECTPTR drawable;

      if (!(movement = Args->Args)) return(ERR_Okay);

      if (movement->OverObject IS Self->RegionID) {
         if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
         else if (Self->State != STATE_INSIDE) Self->State = STATE_ENTERED;
      }
      else {
         if (Self->State IS STATE_EXITED) return(ERR_Okay);
         else Self->State = STATE_EXITED;
      }

      /*** Change the drawable's frame if necessary ***/

      if ((Self->State != STATE_INSIDE) AND (Self->EnterFrame)) {
         if (AccessObject(Self->RegionID, 2000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
            if (Self->State IS STATE_EXITED) {
               SetField(drawable, FID_Frame, FT_LONG, &Self->ExitFrame);
               DelayMsg(AC_Draw, Self->RegionID, NULL);
            }
            else if (Self->State IS STATE_ENTERED) {
               if (!(Self->Flags & INF_DISABLED)) {
                  SetField(drawable, FID_Frame, FT_LONG, &Self->EnterFrame);
                  DelayMsg(AC_Draw, Self->RegionID, NULL);
               }
            }
            ReleaseObject(drawable);
         }
      }

      if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
   }
   else if (Args->ActionID IS AC_Focus) {

   }
   else if (Args->ActionID IS AC_LostFocus) {

   }
   else if (Args->ActionID IS AC_Disable) {
      Self->Flags |= INF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      Self->Flags &= ~INF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else return(ERR_NoSupport);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates the input object's response mechanism.
###END###

*****************************************************************************/

static ERROR INPUT_Activate(struct Input *Self, APTR Void)
{
   STRING script;
   LONG count;
   WORD i;

   DPrintF("~Activate()","[Input:%d]", Self->Head.UniqueID);

   if (Self->Active IS TRUE) {
      DPrintF("@Activate:","Warning - recursion detected");
      StepBack();
      return(ERR_Failed);
   }

   Self->Active = TRUE;

   if (GET_ActionScript(Self, &Self->ActionScript) IS ERR_Okay) {
      if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_Target|TLONG,     Self->RegionID,
         FID_Location|TSTRING, Self->ActionScript,
         TAGEND) IS ERR_Okay) {
         Action(AC_Activate, script, NULL);
         Action(AC_Free, script, NULL);
      }
   }

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];
      if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
         if (Self->Flags & INF_LOCAL) {
            for (i=0; i < count; i++) DelayMsg(AC_Activate, list[i].ObjectID, NULL);
         }
         else for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
      }
   }

   Self->Active = FALSE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Input: ClosingTag
*/

static ERROR INPUT_ClosingTag(struct Input *Self, APTR Void)
{
   if (!(Self->Flags & INF_HIDE)) Action(AC_Show, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Input: DataChannel
*/
                                   
static ERROR INPUT_DataChannel(struct Input *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry *key;
      WORD i;

      if ((key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {
            if (key[i].Flags & KQ_PRESSED) {
               if (key[i].Value IS K_ENTER) {
                  Action(AC_Activate, Self, NULL);
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Turns the input box off.
###END###

*****************************************************************************/

static ERROR INPUT_Disable(struct Input *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is disabled.
   */
   DPrintF("Disable()","[Input:%d]", Self->Head.UniqueID);
   ActionMsg(AC_Disable, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Turns the input box back on if it has previously been disabled.
###END###

*****************************************************************************/

static ERROR INPUT_Enable(struct Input *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is enabled.
   */
   DPrintF("Enable()","[Input:%d]", Self->Head.UniqueID);
   ActionMsg(AC_Enable, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Sets the focus on the input box.
###END###

*****************************************************************************/

static ERROR INPUT_Focus(struct Input *Self, APTR Void)
{
   return(ActionMsg(AC_Focus, Self->RegionID, NULL));
}

/*****************************************************************************
** Input: Free
*/

static ERROR INPUT_Free(struct Input *Self, APTR Void)
{
   if (Self->ActionScriptMID) {
      if (Self->ActionScript) {
         ReleaseMemory(Self->ActionScript);
         Self->ActionScript = NULL;
      }
      FreeMemoryID(Self->ActionScriptMID);
      Self->ActionScriptMID = NULL;
   }

   if (Self->FontID) {
      if (Self->Font) {
         Action(AC_Free, Self->Font, NULL);
         ReleaseObject(Self->Font);
         Self->Font = NULL;
      }
      else ActionMsg(AC_Free, Self->FontID, NULL);
      Self->FontID = NULL;
   }

   if (Self->TextID) {
      if (Self->Text) {
         Action(AC_Free, Self->Text, NULL);
         ReleaseObject(Self->Text);
         Self->Text = NULL;
      }
      else ActionMsg(AC_Free, Self->TextID, NULL);
      Self->TextID = NULL;
   }

   if (Self->RegionID) {
      ActionMsg(AC_Free, Self->RegionID, NULL);
      Self->RegionID = NULL;
   }

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Removes the input box from the display.
###END###

*****************************************************************************/

static ERROR INPUT_Hide(struct Input *Self, APTR Void)
{
   ActionMsg(AC_Hide, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Input: Init()
*/

static ERROR INPUT_Init(struct Input *Self, APTR Void)
{
   struct Render *region;
   OBJECTID owner_id;
   LONG flags;

   /*** Find our parent drawable ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (owner_id) Self->DrawableID = owner_id;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   /*** Initialise the Font ***/

   if (Action(AC_Init, Self->Font, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /*** Initialise the input region ***/

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      SetFields(region, FID_Parent|TLONG, Self->DrawableID,
                        FID_Region|TLONG, TRUE,
                        TAGEND);

      if (!(region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if ((!(region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) OR (!(region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)))) {
            SetFields(region, FID_Height|TLONG, Self->Font->MaxHeight + ((Self->Thickness+2)*2), TAGEND);
         }
      }

      if (Action(AC_Init, region, NULL) IS ERR_Okay) {
         SubscribeActionTags(region, Self,
            AC_Disable,
            AC_Enable,
            AC_Focus,
            AC_LostFocus,
            AC_UserMovement,
            TAGEND);
      }

      ReleaseObject(region);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   /*** Use the base template to create the input graphics ***/

   if (Self->TemplateXML) {
      OBJECTPTR script;
      struct mtGetXMLString getxml;
      struct XMLTag **xmllist, *tag;

      if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            if (tag->Child) {
               if (StrCompare("graphics", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  getxml.Index = tag->Child->Index;
                  getxml.Flags = XMF_INCLUDESIBLINGS;
                  Action(MT_GetXMLString, Self->TemplateXML, &getxml);

                  if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
                     FID_String|TSTRING, getxml.Result,
                     FID_Target|TLONG,   Self->RegionID,
                     TAGEND) IS ERR_Okay) {

                     if (Action(AC_Activate, script, NULL) IS ERR_Okay) {
                        Self->Flags |= INF_NOBKGD;
                     }

                     Action(AC_Free, script, NULL);
                  }

                  FreeMemory(getxml.Result);
                  break;
               }
            }
         }
      }

      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, region, Self, &DrawInput);
      ReleaseObject(region);
   }
   else return(ERR_AccessObject);

   /*** Calculate the width of the text label, if there is one ***/

   if ((Self->LabelWidth < 1) AND (Self->Label[0])) {
      struct mtCalcStringWidth calc;
      calc.String = Self->Label;
      calc.Length = -1;
      Action(MT_CalcStringWidth, Self->Font, &calc);
      Self->LabelWidth = calc.Width + 4;
   }

   /* Initialise the text area that the user will be
   ** able to interact with.
   */

   flags = TXF_EDIT|TXF_ENTERACTIVATE|TXF_ENTERTAB;
   if (Self->Flags & INF_COMMANDLINE)   flags |= TXF_COMMANDLINE;
   if (Self->Flags & INF_FOCUSACTIVATE) flags |= TXF_FOCUSACTIVATE;
   if (Self->Flags & INF_SECRET)        flags |= TXF_SECRET;

   SetFields(Self->Text, FID_Drawable|TLONG,    Self->RegionID,
                         FID_Flags|TLONG,       flags,
                         FID_Face|TSTRING,      Self->Font->Face,
                         FID_Point|TLONG,       Self->Font->Point,
                         FID_XCoord|TLONG,      Self->LabelWidth + Self->Thickness,
                         FID_YCoord|TLONG,      Self->Thickness,
                         FID_XOffset|TLONG,     Self->Thickness,
                         FID_YOffset|TLONG,     Self->Thickness,
                         FID_LineLimit|TLONG,   1,
                         FID_LeftMargin|TLONG,  3,
                         FID_RightMargin|TLONG, 3,
                         FID_Align|TLONG,       ALIGN_VERTICAL,
                         TAGEND);

   if (Action(AC_Init, Self->Text, NULL) IS ERR_Okay) {
      SubscribeActionTags(Self->Text, Self, AC_Activate, TAGEND);
   }
   else return(ObjectError(Self, ERH_Init, ERR_Init));

   if (Self->Flags & INF_SELECTTEXT) ActionTags(MT_SelectTextArea, Self->Text, 0, 0, 20000, 20000);

   if (!(Self->Flags & (INF_SUNKEN|INF_RAISED))) Self->Flags |= INF_SUNKEN;

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the input box to the back of the display area.
###END###

*****************************************************************************/

static ERROR INPUT_MoveToBack(struct Input *Self, APTR Void)
{
   ActionMsg(AC_MoveToBack, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToFront
Short: Moves the input box to the front of the display area.
###END###

*****************************************************************************/

static ERROR INPUT_MoveToFront(struct Input *Self, APTR Void)
{
   ActionMsg(AC_MoveToFront, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Input: NewObject()
*/

static ERROR INPUT_NewObject(struct Input *Self, APTR Void)
{
   OBJECTID colours_id;
   OBJECTPTR colours;

   if (NewObject(ID_RENDER, NF_CHILD|Self->Head.ObjectFlags, NULL, &Self->RegionID) IS ERR_Okay) {
      if (NewObject(ID_FONT, NF_CHILD|Self->Head.ObjectFlags, &Self->Font, &Self->FontID) IS ERR_Okay) {
         SetField(Self->Font, FID_Face, FT_STRING, "sans serif");

         if (NewObject(ID_TEXT, NF_CHILD, &Self->Text, &Self->TextID) IS ERR_Okay) {
            Self->ExitFrame = 1;
            Self->ReleaseFrame = 1;

            Self->Flags |= INF_SUNKEN;

            /*** Internal colour ***/
            Self->ColourRGB.Red   = 0;
            Self->ColourRGB.Green = 255;
            Self->ColourRGB.Blue  = 255;
            Self->ColourRGB.Alpha = 255;

            /*** Shadow colour ***/
            Self->ShadowRGB.Red   = 100;
            Self->ShadowRGB.Green = 100;
            Self->ShadowRGB.Blue  = 100;
            Self->ShadowRGB.Alpha = 255;

            /*** Highlight colour ***/
            Self->HighlightRGB.Red   = 255;
            Self->HighlightRGB.Green = 255;
            Self->HighlightRGB.Blue  = 255;
            Self->HighlightRGB.Alpha = 255;

            /*** Try to use colours from the user's preferences ***/

            if (FastFindObject("gl_colours", ID_VARIABLE, &colours_id, 1, NULL) IS ERR_Okay) {
               if (AccessObject(colours_id, 4000, &colours) IS ERR_Okay) {
                  UBYTE buffer[100];

                  if (GetFieldVariable(colours, "background", buffer, sizeof(buffer)) IS ERR_Okay) {
                     SET_Colour(Self, buffer);
                  }

                  if (GetFieldVariable(colours, "highlight", buffer, sizeof(buffer)) IS ERR_Okay) {
                     SET_Highlight(Self, buffer);
                  }

                  if (GetFieldVariable(colours, "shadow", buffer, sizeof(buffer)) IS ERR_Okay) {
                     SET_Shadow(Self, buffer);
                  }

                  ReleaseObject(colours);
               }
            }

            Self->Thickness = 1;

            SET_Template(Self, "templates:inputdef.xml");

            return(ERR_Okay);
         }
         else return(ERR_NewObject);
      }
      else return(ERR_NewObject);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Changes the size and position of the input box.
###END###

*****************************************************************************/

static ERROR INPUT_Redimension(struct Input *Self, struct acRedimension *Args)
{
   return(ActionMsg(AC_Redimension, Self->RegionID, Args));
}

/*****************************************************************************
** Input: ReleaseObject
*/

static ERROR INPUT_ReleaseObject(struct Input *Self, APTR Void)
{
   if (Self->Font) {
      ReleaseObject(Self->Font);
      Self->Font = NULL;
   }

   if (Self->Text) {
      ReleaseObject(Self->Text);
      Self->Text = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Alters the size of the input box.
###END###

*****************************************************************************/

static ERROR INPUT_Resize(struct Input *Self, struct acResize *Args)
{
   return(ActionMsg(AC_Resize, Self->RegionID, Args));
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Puts the input box on display.
###END###

*****************************************************************************/

static ERROR INPUT_Show(struct Input *Self, APTR Void)
{
   ActionMsg(AC_Show, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ActionScript
Short:  Script to be executed when the input box is clicked.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When an input box is activated, you can elect to run a script that
performs a responsive action.  The ActionScript string will typically refer
to a file location, although the special STRING: assignment followed with
the script execution text (e.g. DML code) can also be useful for optimising
the execution of simple scripts.</p>

###END###

*****************************************************************************/

static ERROR GET_ActionScript(struct Input *Self, STRING *Value)
{
   if (Self->ActionScript) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else if (!Self->ActionScriptMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ActionScriptMID, MEM_READ, (void **)&Self->ActionScript) IS ERR_Okay) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_ActionScript(struct Input *Self, STRING Value)
{
   LONG i;

   if (Self->ActionScript)    { ReleaseMemory(Self->ActionScript);   Self->ActionScript = NULL; }
   if (Self->ActionScriptMID) { FreeMemoryID(Self->ActionScriptMID); Self->ActionScriptMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->ActionScript, &Self->ActionScriptMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->ActionScript[i] = Value[i];
         Self->ActionScript[i] = 0;
      }
      else return(ObjectError((OBJECTPTR)Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Bottom
Short:  The bottom coordinate of the input box.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The bottom coordinate of the input box (calculated as YCoord + Height) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Bottom(struct Input *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Bottom, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for setting the colour inside of the input box.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The colour inside of an input box is declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

###SEE ALSO###
Fields: Highlight, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Colour(struct Input *Self, STRING Colour)
{
   if (Colour) {
      StrReadColour(Colour, &Self->ColourRGB.Red, &Self->ColourRGB.Green, &Self->ColourRGB.Blue, &Self->ColourRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   FocusFrame
Short:  The graphics frame to display when the input box has the focus.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user
focusses on the input box.  By default this field is initially set to
zero, which has no effect on the drawable frame.  When the user leaves
the input box, it will revert to the frame indicated by the ReleaseFrame
field.</p>

###SEE ALSO###
Field: ReleaseFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Defines the internal colour of the input box in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The colour inside of the input box can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct Input *Self, struct RGB *Value)
{
   Value->Red   = Self->ColourRGB.Red;
   Value->Green = Self->ColourRGB.Green;
   Value->Blue  = Self->ColourRGB.Blue;
   Value->Alpha = Self->ColourRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR GET_Disable(struct Input *Self, LONG *Value)
{
   if (Self->Flags & INF_DISABLED) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Disable(struct Input *Self, LONG *Value)
{
   if (*Value IS TRUE) Action(AC_Disable, Self, NULL);
   else if (*Value IS FALSE) Action(AC_Enable, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  The drawable that will contain the input graphic.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that will contain the input graphic is set here.  If this
field is not set prior to initialisation, the input will attempt to scan
for the correct drawable by analysing its parents until it finds a suitable
candidate.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   EnterFrame
Short:  The graphics frame to display when the user's pointer enters the input area.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user's
pointer enters the input area.  By default this field is initially set to
zero, which has no effect on the drawable frame.</p>

###SEE ALSO###
Field: ExitFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ExitFrame
Short:  The graphics frame to display when the user's pointer leaves the input area.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user's pointer
leaves the input area.  This field is unused if the EnterFrame field has
not been set.</p>

###SEE ALSO###
Field: EnterFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a input object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">COMMANDLINE</td><td>Prepares the input box for 'command-line' mode.  This option is typically used when the input box needs to remain persistent and is linked to an output system, such as a shell.</td></tr>
<tr><td width="10%">DISABLED</td><td>This readable flag is automatically set when the input is disabled.  You may set it on initialisation if the input is to be disabled on creation.</td></tr>
<tr><td width="10%">FOCUSACTIVATE</td><td>If the input box loses the focus, the presence of this option will force it to automtically activate itself (as if the user had pressed the enter key).</td></tr>
<tr><td width="10%">HIDE</td><td>This flag may be set on initialisation in order to hide the input on its creation (applies to script usage only).</td></tr>
<tr><td width="10%">LOCAL</td><td>When the input is activated, any child objects belonging to the input will be executed immediately within the input's execution process, rather than through the messaging system.</td></tr>
<tr><td width="10%">RAISED</td><td>(Read-Only) Draws the input box with the raised effect.</td></tr>
<tr><td width="10%">SECRET</td><td>Use the SECRET option when the input box is being used to enter secret user information, such as a password.</td></tr>
<tr><td width="10%">SUNKEN</td><td>(Read-Only) Draws the input box using the sunken effect.  This is the default.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   FocusActivate
Short:  If set to TRUE, the input box will automatically activate itself if the user leaves the input area.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>If this field is set to TRUE, the input box will automatically activate
itself if the user enters the input area and then leaves.  The activation
will occur regardless of wether or not the user edited the content of the
input box.  The effect is the same as if the user had proactively pressed
the enter key.</p>

###END###
*****************************************************************************/

static ERROR GET_FocusActivate(struct Input *Self, LONG *Value)
{
   if (Self->Flags & INF_FOCUSACTIVATE) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_FocusActivate(struct Input *Self, LONG *Value)
{
   if (*Value IS TRUE) Self->Flags |= INF_FOCUSACTIVATE;
   else Self->Flags &= ~INF_FOCUSACTIVATE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of an input box.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>An input box can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));
}

static ERROR SET_Height(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Height, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the border highlight.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border highlight of an input box is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct Input *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines border highlight of the input box, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border highlight of the input box can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct Input *Self, struct RGB *Value)
{
   Value->Red   = Self->HighlightRGB.Red;
   Value->Green = Self->HighlightRGB.Green;
   Value->Blue  = Self->HighlightRGB.Blue;
   Value->Alpha = Self->HighlightRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Label
Short:  The label is a string displayed to the left of the input area.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>A label can be drawn next to the input area by setting the Label field.
The label should be a short, descriptive string of one or two words.  It
is common practice for the label to be followed with a colon character.</p>

###SEE ALSO###
Field: LabelWidth

###END###

*****************************************************************************/

static ERROR GET_Label(struct Input *Self, STRING *Value)
{
   *Value = Self->Label;
   return(ERR_Okay);
}

static ERROR SET_Label(struct Input *Self, STRING Value)
{
   WORD i;
   if (Value) {
      for (i=0; (Value[i]) AND (i < sizeof(Self->Label)-1); i++) Self->Label[i] = Value[i];
      Self->Label[i] = 0;
   }
   else Self->Label[0] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   LabelWidth
Short:  The width of the input label.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>If a label has been set for an input box, its width may be read and
adjusted at any time via the LabelWidth field.  The input area will be
arranged so that it immediately follows the width defined for the text
label.  If you define a width that is too short for the text that is to be
printed in the label, the text will be trimmed to fit the defined area.</p>

<p>If you specify a label without setting the label width, the correct width
will be automatically calculated for you on initialisation.</p>

###END###

*****************************************************************************/

static ERROR SET_LabelWidth(struct Input *Self, LONG *Value)
{
   LONG x;

   Self->LabelWidth = *Value;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      x = Self->LabelWidth + Self->Thickness;
      SetField(Self->Text, FID_XCoord, FT_LONG, &x);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Raised
Short:  If set to TRUE the input box will appear to be raised into the foreground.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of an input object then you
will need to decide whether or not the box should be given a sunken or raised
effect when it is drawn.  To give it a raised effect you will need to set this
field to TRUE, if not then you should set the Sunken field.</p>

###SEE ALSO###
Field: Sunken

###END###

*****************************************************************************/

static ERROR GET_Raised(struct Input *Self, LONG *Value)
{
   if (Self->Flags & INF_SUNKEN) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Raised(struct Input *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & INF_RAISED) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~INF_SUNKEN) | INF_RAISED;
   }
   else Self->Flags &= ~INF_RAISED;

   if (Self->Flags & INF_ACTIVEDRAW) ActionMsg(AC_Draw, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Region
Short:  The drawable that represents the input box is referenced through this field.
Type:   OBJECTID
Status: Read

###DESCRIPTION###
<p>The rendering area that represents the input display can be accessed
through this field.  For further information, refer to the Render class.  Note
that talking to the drawable directly can have adverse effects on the input
control system.  Where possible, all communication should be limited to the
input object itself.</p>

###SEE ALSO###
Class: Render

###END###

*****************************************************************************/

static ERROR SET_Region(struct Input *Self, LONG *Value)
{
   OBJECTPTR drawable;

   /* NOTE: For backwards compatibility with the Render class, the region
   ** can be set to a value of TRUE to define the input as a simple drawable
   ** region.
   */

   if ((*Value IS FALSE) OR (*Value IS TRUE)) {
      if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
         SetField(drawable, FID_Region, FT_LONG, Value);
         ReleaseObject(drawable);
         return(ERR_Okay);
      }
      else return(ERR_AccessObject);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   ReleaseFrame
Short:  The graphics frame to display when the input box loses the focus.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If the FocusFrame field has been set, you may want to match that value
by indicating the frame that should be used when the click is released.
By default, the value in this field will initially be set to 1.  This field
is unused if the FocusFrame field has not been set.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Right
Short:  The right-most coordinate of the input box.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The right-most coordinate of the input box (calculated as XCoord + Width) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Right(struct Input *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Right, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the input box shadow.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border shadow of an input box is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Highlight, Thickness

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct Input *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->ShadowRGB.Red, &Self->ShadowRGB.Green, &Self->ShadowRGB.Blue, &Self->ShadowRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines the border shadow of the input box, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border shadow of the input box can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct Input *Self, struct RGB *Value)
{
   Value->Red   = Self->ShadowRGB.Red;
   Value->Green = Self->ShadowRGB.Green;
   Value->Blue  = Self->ShadowRGB.Blue;
   Value->Alpha = Self->ShadowRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   String
Short:  The string that is to be printed inside the input box is declared here.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The string that you would like to be displayed in the input box is
specified in this field.  The string must be in UTF-8 format and may not
contain line feeds.  You can read this field at any time to determine what
the user has entered in the input box.</p>

<p>If the string is changed after initialisation, the input box will be redrawn
to show the updated text.</p>

###END###

*****************************************************************************/

static ERROR GET_String(struct Input *Self, STRING *Value)
{
   STRING str;
   ERROR error;

   if ((error = GetField(Self->Text, FID_String, FT_STRING, &str)) IS ERR_Okay) {
      *Value = str;
      return(ERR_Okay);
   }
   else return(error);
}

static ERROR SET_String(struct Input *Self, STRING Value)
{
   if (SetField(Self->Text, FID_String, FT_STRING, Value) IS ERR_Okay) {
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   Sunken
Short:  Set to TRUE to make the input box appear to sink into the background.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of an input object then
you will need to decide whether or not the box should be given a sunken or
raised effect when it is drawn.  To give it a sunken effect you will need to
set this field to TRUE, if not then you should set the Raised field.</p>

###SEE ALSO###
Field: Raised

###END###

*****************************************************************************/

static ERROR GET_Sunken(struct Input *Self, LONG *Value)
{
   if (Self->Flags & INF_SUNKEN) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Sunken(struct Input *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & INF_SUNKEN) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~INF_RAISED) | INF_SUNKEN;
   }
   else Self->Flags &= ~INF_SUNKEN;

   if (Self->Flags & INF_ACTIVEDRAW) ActionMsg(AC_Draw, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TabFocus
Short:  Setting this field to a valid TabFocus object will cause the input to add itself to the tab list.
Type:   OBJECTID
Status: Set

###DESCRIPTION###
<p>The TabFocus field provides a convenient way of adding the input to a
TabFocus object, so that it can be focussed on via the tab key.  Simply set
this field to the ID of the TabFocus object that is managing the tab-list
for the application window.</p>

###SEE ALSO###
Class: TabFocus

###END###

*****************************************************************************/

static ERROR SET_TabFocus(struct Input *Self, OBJECTID *Value)
{
   OBJECTPTR tabfocus;

   if (AccessObject(*Value, 5000, &tabfocus) IS ERR_Okay) {
      if (tabfocus->ClassID IS ID_TABFOCUS) {
         ActionTags(MT_AddTabObject, tabfocus, Self->Head.UniqueID); //RegionID);
      }
      ReleaseObject(tabfocus);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Template
Short:  Defines the makeup of the input using a pre-defined template.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>Setting the Template field allows a input defintion to be loaded from a
pre-defined template file.  The template may contain settings for all of the
field values of a input object, as well as a customised graphics script
for building the look and feel of the input.  The template must be provided
as an XML file that can contain two possible root tags - 'value' and 'graphics'.
The value tag contains a list of field values for setting against the input,
while the graphics tag contains the graphical makeup of the input.  The
following example illustrates how a customised input can be created:</p>

<pre>
&lt;?xml version="1.0"?&gt;

&lt;values&gt;
  &lt;exitframe value="1"/&gt;
  &lt;focusframe value="2"/&gt;
  &lt;enterframe value="3"/&gt;
  &lt;releaseframe value="1"/&gt;
&lt;/values&gt;

&lt;graphics&gt;
  &lt;box frame="1" boxes="(s0,0,!0,!0) (s1,1,!1,!1)" colour="230,230,230"
    highlight(0)="120,120,120" highlight(1)="255,255,255" shadow(0)="80,80,80"
    shadow(1)="200,200,200"/&gt;
&lt;/graphics&gt;
</pre>

###END###

*****************************************************************************/

static ERROR SET_Template(struct Input *Self, STRING Value)
{
   struct XMLTag **xmllist, *tag, *child;
   UBYTE value[300];
   WORD i;

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AnalyseLocation(Value, NULL) IS ERR_Okay) {
      if (CreateObject(ID_XML, NF_CHILD, &Self->TemplateXML, NULL,
         FID_Location|TSTRING, Value,
         TAGEND) IS ERR_Okay) {

         if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
            for (tag=xmllist[0]; tag; tag=tag->Next) {
               if (tag->Child) {
                  if (StrCompare("values", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     for (child=tag->Child; child; child=child->Next) {
                        if (child->TotalAttributes > 1) {
                           for (i=0; (i < sizeof(value)-1) AND (child->Attributes[1].Value[i]); i++) value[i] = child->Attributes[1].Value[i];
                           value[i] = 0;
                           StrTranslate(value, sizeof(value), NULL);
                           SetFieldVariable(Self, child->Attributes[0].Name, value);
                        }
                     }
                     break;
                  }
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Thickness
Short:  The thickness of the input border.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the thickness of the input border, expressed in
pixels.  By default the thickness is set to a value of 1.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a input.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A input can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));
}

static ERROR SET_Width(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Width, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a input.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a input can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetX, ERR_AccessObject));
}

static ERROR SET_X(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a input.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the input will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the Input is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the input will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - InputWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_AccessObject));
}

static ERROR SET_XOffset(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a input.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a Input can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

}

static ERROR SET_Y(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a input.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the input will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the input is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the input will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - InputHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_AccessObject));
}

static ERROR SET_YOffset(struct Input *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************
** Internal: DrawInput()
*/

static void DrawInput(struct Input *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct Font *font;
   struct mtDrawRectangle rect;
   ULONG highlight, shadow;
   WORD i, x, width;
   UBYTE salpha, halpha;

   if (!(Self->Flags & INF_NOBKGD)) {
      Bitmap->Opacity = Self->ColourRGB.Alpha;
      rect.XCoord = Self->LabelWidth;
      rect.YCoord = 0;
      rect.Width  = Drawable->Width - Self->LabelWidth;
      rect.Height = Drawable->Height;
      rect.Fill   = TRUE;
      rect.Colour = PackPixel(Bitmap, Self->ColourRGB.Red, Self->ColourRGB.Green, Self->ColourRGB.Blue);
      Action(MT_DrawRectangle, Bitmap, &rect);

      /*** Draw the borders around the rectangular area ***/

      if (Self->Flags & INF_SUNKEN) {
         /*** Reverse the border definitions in sunken mode ***/
         highlight = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
         halpha = Self->ShadowRGB.Alpha;
         shadow = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);
         salpha = Self->HighlightRGB.Alpha;
      }
      else {
         shadow = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
         salpha = Self->ShadowRGB.Alpha;
         highlight = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);
         halpha = Self->HighlightRGB.Alpha;
      }

      x = Self->LabelWidth;
      width = Drawable->Width - Self->LabelWidth;
      if (Self->Thickness > 0) {
         for (i=0; i < Self->Thickness; i++) {
            /*** Top, Bottom ***/

            Bitmap->Opacity = halpha;
            ActionTags(MT_DrawRectangle, Bitmap, x+i, i, width-i-i, 1, highlight, TRUE);
            Bitmap->Opacity = salpha;
            ActionTags(MT_DrawRectangle, Bitmap, x+i, Drawable->Height-i-1, width-i-i, 1, shadow, TRUE);

            /*** Left, Right ***/

            Bitmap->Opacity = halpha;
            ActionTags(MT_DrawRectangle, Bitmap, x+i,         i+1, 1, Drawable->Height-i-i-2, highlight, TRUE);
            Bitmap->Opacity = salpha;
            ActionTags(MT_DrawRectangle, Bitmap, x+width-i-1, i+1, 1, Drawable->Height-i-i-2, shadow, TRUE);
         }
      }

      Bitmap->Opacity = 255;
   }

   if (Self->Label[0]) {
      font = Self->Font;

      SetField(font, FID_Bitmap, FT_POINTER, Bitmap);
      SetField(font, FID_String, FT_STRING, Self->Label);

      if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 25, TAGEND);

      font->XCoord = 0;
      font->YCoord = 0;
      font->AlignFlags |= ALIGN_VERTICAL;
      font->AlignWidth  = Drawable->Width;
      font->AlignHeight = Drawable->Height;
      Action(AC_Draw, font, NULL);

      if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 0, TAGEND);
   }
}

